from __future__ import annotations

from app.data.database.components import ComponentType
from app.data.database.database import DB
from app.data.database.item_components import ItemComponent, ItemTags
from app.engine import (action, banner, combat_calcs, engine, equations,
                        image_mods, item_funcs, item_system, skill_system,
                        target_system)
from app.engine.game_state import game
from app.engine.objects.unit import UnitObject
from app.utilities import utils, static_random
from app.data.database.difficulty_modes import RNGOption
from app.engine.combat import playback as pb
from app.engine.movement import movement_funcs
import logging


class DoNothing(ItemComponent):
    nid = 'do_nothing'
    desc = 'does nothing'
    tag = ItemTags.CUSTOM

    expose = ComponentType.Int
    value = 1
    
class WeaponTypeTag(ItemComponent):
    nid = 'weapon_type_tag'
    desc = "Defines the type of weapon. Anybody can use it though"
    tag = ItemTags.CUSTOM

    expose = ComponentType.WeaponType

    def weapon_type(self, unit, item):
        return self.value

class EvalEnemyBlastAOE(ItemComponent):
    nid = 'eval_smartblast_aoe'
    desc = "Grants EVAL Enemy AoE range."
    tag = ItemTags.CUSTOM

    expose = ComponentType.String

    def _get_power(self, unit) -> int:
        from app.engine import evaluate
        try:
            base_power = int(evaluate.evaluate(self.value, unit))
        except Exception as e:
            logging.error("Couldn't evaluate %s conditional (%s)", self.value, e)
            base_power = 0
        empowered_splash = skill_system.empower_splash(unit)
        return base_power + 1 + empowered_splash

    def splash(self, unit, item, position) -> tuple:
        ranges = set(range(self._get_power(unit)))
        splash = game.target_system.find_manhattan_spheres(ranges, position[0], position[1])
        splash = {pos for pos in splash if game.board.check_bounds(pos)}
        from app.engine import item_system, skill_system
        if item_system.is_spell(unit, item):
            # spell blast
            splash = [game.board.get_unit(s) for s in splash]
            splash = [s.position for s in splash if s and skill_system.check_enemy(unit, s)]
            return None, splash
        else:
            # regular blast
            splash = [game.board.get_unit(s) for s in splash if s != position]
            splash = [s.position for s in splash if s and skill_system.check_enemy(unit, s)]
            return position if game.board.get_unit(position) else None, splash

    def splash_positions(self, unit, item, position) -> set:
        from app.engine import skill_system
        ranges = set(range(self._get_power(unit)))
        splash = game.target_system.find_manhattan_spheres(ranges, position[0], position[1])
        splash = {pos for pos in splash if game.tilemap.check_bounds(pos)}
        # Doesn't highlight allies positions
        splash = {pos for pos in splash if not game.board.get_unit(pos) or skill_system.check_enemy(unit, game.board.get_unit(pos))}
        return splash

class EnemySplashRestrict(ItemComponent):
    nid = 'enemy_splash_restrict'
    desc = "To be used alongside enemy blast AoE"
    tag = ItemTags.CUSTOM
        
    def target_restrict(self, unit, item, def_pos, splash) -> bool:
        # Restricts target based on whether any unit is an enemy
        defender = game.board.get_unit(def_pos)
        if defender and skill_system.check_enemy(unit, defender):
            return True
        for s_pos in splash:
            s = game.board.get_unit(s_pos)
            if s and skill_system.check_enemy(unit, s):
                return True
        return False
        
class EvalDamage(ItemComponent):
    nid = 'eval_damage'
    desc = "Item does damage on hit. Damage is evaluated."
    tag = ItemTags.CUSTOM

    expose = ComponentType.String
    value = 0

    def damage(self, unit, item):
        from app.engine import evaluate
        try:
            return int(evaluate.evaluate(self.value, unit, local_args={'item': item}))
        except Exception as e:
            logging.error("EVAL DAMAGE: Couldn't evaluate %s conditional (%s)", self.value, e)
            return 0

    def target_restrict(self, unit, item, def_pos, splash) -> bool:
        # Restricts target based on whether any unit is an enemy
        defender = game.board.get_unit(def_pos)
        if defender and skill_system.check_enemy(unit, defender):
            return True
        for s_pos in splash:
            s = game.board.get_unit(s_pos)
            if s and skill_system.check_enemy(unit, s):
                return True
        return False

    def on_hit(self, actions, playback, unit, item, target, item2, target_pos, mode, attack_info):
        playback_nids = [brush.nid for brush in playback]
        if 'attacker_partner_phase' in playback_nids or 'defender_partner_phase' in playback_nids:
            damage = combat_calcs.compute_assist_damage(unit, target, item, target.get_weapon(), mode, attack_info)
        else:
            damage = combat_calcs.compute_damage(unit, target, item, target.get_weapon(), mode, attack_info)

        # Reduce damage if in Grandmaster Mode
        if game.mode.rng_choice == RNGOption.GRANDMASTER:
            hit = utils.clamp(combat_calcs.compute_hit(unit, target, item, target.get_weapon(), mode, attack_info), 0, 100)
            damage = int(damage * float(hit) / 100)

        true_damage = min(damage, target.get_hp())
        actions.append(action.ChangeHP(target, -damage))

        # For animation
        playback.append(pb.DamageHit(unit, item, target, damage, true_damage))
        if damage == 0:
            playback.append(pb.HitSound('No Damage'))
            playback.append(pb.HitAnim('MapNoDamage', target))

    def on_glancing_hit(self, actions, playback, unit, item, target, item2, target_pos, mode, attack_info):
        playback_nids = [brush.nid for brush in playback]
        if 'attacker_partner_phase' in playback_nids or 'defender_partner_phase' in playback_nids:
            damage = combat_calcs.compute_assist_damage(unit, target, item, target.get_weapon(), mode, attack_info)
        else:
            damage = combat_calcs.compute_damage(unit, target, item, target.get_weapon(), mode, attack_info)

        # Reduce damage if in Grandmaster Mode
        if game.mode.rng_choice == RNGOption.GRANDMASTER:
            hit = utils.clamp(combat_calcs.compute_hit(unit, target, item, target.get_weapon(), mode, attack_info), 0, 100)
            damage = int(damage * float(hit) / 100)

        damage //= 2  # Because glancing hit

        true_damage = min(damage, target.get_hp())
        actions.append(action.ChangeHP(target, -damage))

        # For animation
        playback.append(pb.DamageHit(unit, item, target, damage, true_damage))
        if damage == 0:
            playback.append(pb.HitAnim('MapNoDamage', target))
        else:
            playback.append(pb.HitAnim('MapGlancingHit', target))

    def on_crit(self, actions, playback, unit, item, target, item2, target_pos, mode, attack_info):
        playback_nids = [brush.nid for brush in playback]
        if 'attacker_partner_phase' in playback_nids or 'defender_partner_phase' in playback_nids:
            damage = combat_calcs.compute_assist_damage(unit, target, item, target.get_weapon(), mode, attack_info, crit=True)
        else:
            damage = combat_calcs.compute_damage(unit, target, item, target.get_weapon(), mode, attack_info, crit=True)

        # Reduce damage if in Grandmaster Mode (although crit doesn't make much sense with Grandmaster mode)
        if game.mode.rng_choice == RNGOption.GRANDMASTER:
            hit = utils.clamp(combat_calcs.compute_hit(unit, target, item, target.get_weapon(), mode, attack_info), 0, 100)
            damage = int(damage * float(hit) / 100)

        true_damage = min(damage, target.get_hp())
        actions.append(action.ChangeHP(target, -damage))

        playback.append(pb.DamageCrit(unit, item, target, damage, true_damage))
        if damage == 0:
            playback.append(pb.HitSound('No Damage'))
            playback.append(pb.HitAnim('MapNoDamage', target))

def ai_status_priority_buff(unit, target, item, move, status_nid) -> float:
    if target and status_nid not in [skill.nid for skill in target.skills]:
        accuracy_term = utils.clamp(combat_calcs.compute_hit(unit, target, item, target.get_weapon(), "attack", (0, 0))/100., 0, 1)
        num_attacks = combat_calcs.outspeed(unit, target, item, target.get_weapon(), "attack", (0, 0))
        accuracy_term *= num_attacks
        # Tries to maximize distance from target
        distance_term = 0.01 * utils.calculate_distance(move, target.position)
        if skill_system.check_enemy(unit, target):
            return -0.5 * accuracy_term + distance_term
        else:
            return 0.5 * accuracy_term
    return 0

class EnemyWideSwordAOE(ItemComponent):
    nid = 'enemy_widesword_aoe'
    desc = "Widesword"
    tag = ItemTags.CUSTOM

    def splash(self, unit, item, position) -> tuple:
        from app.engine import skill_system
        pos = unit.position
        if position[0] == pos[0] + 1:
            all_positions = {(pos[0] + 1, pos[1] - 1),
                             (pos[0] + 1, pos[1]),
                             (pos[0] + 1, pos[1] + 1)}
        elif position[0] == pos[0] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0] - 1, pos[1]),
                             (pos[0] - 1, pos[1] + 1)}
        elif position[1] == pos[1] + 1:
            all_positions = {(pos[0] - 1, pos[1] + 1),
                             (pos[0], pos[1] + 1),
                             (pos[0] + 1, pos[1] + 1)}
        elif position[1] == pos[1] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0], pos[1] - 1),
                             (pos[0] + 1, pos[1] - 1)}

        all_positions = {pos for pos in all_positions if game.tilemap.check_bounds(pos)}
        all_positions.discard(position)
        splash = all_positions
        splash = [game.board.get_unit(pos) for pos in splash]
        splash = [s.position for s in splash if s and skill_system.check_enemy(unit, s)]
        main_target = position if game.board.get_unit(position) else None
        return main_target, splash

    def splash_positions(self, unit, item, position) -> set:
        from app.engine import skill_system
        pos = unit.position
        if position[0] == pos[0] + 1:
            all_positions = {(pos[0] + 1, pos[1] - 1),
                             (pos[0] + 1, pos[1]),
                             (pos[0] + 1, pos[1] + 1)}
        elif position[0] == pos[0] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0] - 1, pos[1]),
                             (pos[0] - 1, pos[1] + 1)}
        elif position[1] == pos[1] + 1:
            all_positions = {(pos[0] - 1, pos[1] + 1),
                             (pos[0], pos[1] + 1),
                             (pos[0] + 1, pos[1] + 1)}
        elif position[1] == pos[1] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0], pos[1] - 1),
                             (pos[0] + 1, pos[1] - 1)}

        all_positions = {pos for pos in all_positions if game.tilemap.check_bounds(pos)}
        all_positions.discard(position)
        splash = all_positions
        # Doesn't highlight allies positions
        splash = {pos for pos in splash if not game.board.get_unit(pos) or skill_system.check_enemy(unit, game.board.get_unit(pos))}
        return splash

#works, but feels ineffecient
class EnemyLifeSwordAOE(ItemComponent):
    nid = 'enemy_lifesword_aoe'
    desc = "LifeSword"
    tag = ItemTags.CUSTOM

    def splash(self, unit, item, position) -> tuple:
        from app.engine import skill_system
        pos = unit.position
        if position[0] == pos[0] + 1:
            all_positions = {(pos[0] + 1, pos[1] - 1),
                             (pos[0] + 1, pos[1]),
                             (pos[0] + 1, pos[1] + 1),
                             (pos[0] + 2, pos[1] - 1),
                             (pos[0] + 2, pos[1]),
                             (pos[0] + 2, pos[1] + 1)}
        elif position[0] == pos[0] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0] - 1, pos[1]),
                             (pos[0] - 1, pos[1] + 1),
                             (pos[0] - 2, pos[1] - 1),
                             (pos[0] - 2, pos[1]),
                             (pos[0] - 2, pos[1] + 1)}
        elif position[1] == pos[1] + 1:
            all_positions = {(pos[0] - 1, pos[1] + 1),
                             (pos[0] - 1, pos[1] + 2),
                             (pos[0], pos[1] + 1),
                             (pos[0], pos[1] + 2),
                             (pos[0] + 1, pos[1] + 1),
                             (pos[0] + 1, pos[1] + 2)}
        elif position[1] == pos[1] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0], pos[1] - 1),
                             (pos[0] + 1, pos[1] - 1),
                             (pos[0] - 1, pos[1] - 2),
                             (pos[0], pos[1] - 2),
                             (pos[0] + 1, pos[1] - 2)}

        all_positions = {pos for pos in all_positions if game.tilemap.check_bounds(pos)}
        all_positions.discard(position)
        splash = all_positions
        splash = [game.board.get_unit(pos) for pos in splash]
        splash = [s.position for s in splash if s and skill_system.check_enemy(unit, s)]
        main_target = position if game.board.get_unit(position) else None
        return main_target, splash

    def splash_positions(self, unit, item, position) -> set:
        from app.engine import skill_system
        pos = unit.position
        if position[0] == pos[0] + 1:
            all_positions = {(pos[0] + 1, pos[1] - 1),
                             (pos[0] + 1, pos[1]),
                             (pos[0] + 1, pos[1] + 1),
                             (pos[0] + 2, pos[1] - 1),
                             (pos[0] + 2, pos[1]),
                             (pos[0] + 2, pos[1] + 1)}
        elif position[0] == pos[0] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0] - 1, pos[1]),
                             (pos[0] - 1, pos[1] + 1),
                             (pos[0] - 2, pos[1] - 1),
                             (pos[0] - 2, pos[1]),
                             (pos[0] - 2, pos[1] + 1)}
        elif position[1] == pos[1] + 1:
            all_positions = {(pos[0] - 1, pos[1] + 1),
                             (pos[0] - 1, pos[1] + 2),
                             (pos[0], pos[1] + 1),
                             (pos[0], pos[1] + 2),
                             (pos[0] + 1, pos[1] + 1),
                             (pos[0] + 1, pos[1] + 2)}
        elif position[1] == pos[1] - 1:
            all_positions = {(pos[0] - 1, pos[1] - 1),
                             (pos[0], pos[1] - 1),
                             (pos[0] + 1, pos[1] - 1),
                             (pos[0] - 1, pos[1] - 2),
                             (pos[0], pos[1] - 2),
                             (pos[0] + 1, pos[1] - 2)}

        all_positions = {pos for pos in all_positions if game.tilemap.check_bounds(pos)}
        all_positions.discard(position)
        splash = all_positions
        # Doesn't highlight allies positions
        splash = {pos for pos in splash if not game.board.get_unit(pos) or skill_system.check_enemy(unit, game.board.get_unit(pos))}
        return splash