from __future__ import annotations

from app.data.database.components import ComponentType
from app.data.database.database import DB
from app.data.database.skill_components import SkillComponent, SkillTags
from app.engine import (action, banner, combat_calcs, engine, equations,
                        image_mods, item_funcs, item_system, skill_system,
                        target_system)
from app.engine.game_state import game
from app.engine.objects.unit import UnitObject
from app.utilities import utils, static_random

from app.engine.combat import playback as pb
from app.utilities.enums import Strike
import logging


class DoNothing(SkillComponent):
    nid = 'do_nothing'
    desc = 'does nothing'
    tag = SkillTags.CUSTOM

    expose = ComponentType.Int
    value = 1

class RegenerationFormula(SkillComponent):
    nid = 'regeneration_formula'
    desc = "Unit restores %% of HP at beginning of turn"
    tag = SkillTags.CUSTOM

    expose = ComponentType.Equation
    value = 'MAGIC_DAMAGE'

    def on_upkeep(self, actions, playback, unit):
        max_hp = equations.parser.hitpoints(unit)
        if unit.get_hp() < max_hp:
            #hp_change = int(max_hp * self.value)
            hp_change = equations.parser.magic_damage(unit)
            actions.append(action.ChangeHP(unit, self.value))
            # Playback
            playback.append(pb.HitSound('MapHeal'))
            playback.append(pb.DamageNumbers(unit, -self.value))
            if self.value >= 30:
                name = 'MapBigHealTrans'
            elif self.value >= 15:
                name = 'MapMediumHealTrans'
            else:
                name = 'MapSmallHealTrans'
            playback.append(pb.CastAnim(name))            

class FullMiracle(SkillComponent):
    nid = 'full_miracle'
    desc = "Unit will not die after combat, but will instead be resurrected with full hp"
    tag = SkillTags.CUSTOM

    def cleanup_combat(self, playback, unit, item, target, item2, mode):
        if unit.get_hp() <= 0:
            action.do(action.SetHP(unit, unit.get_max_hp()))
            game.death.miracle(unit)
            action.do(action.TriggerCharge(unit, self.skill))
            
class UpkeepSkillGain(SkillComponent):
    nid = 'upkeep_skill_gain'
    desc = "Grants the designated skill at upkeep"
    tag = SkillTags.CUSTOM

    expose = ComponentType.Skill

    def on_upkeep(self, actions, playback, unit):
        action.do(action.AddSkill(unit, self.value))

class LostOnTakeHit(SkillComponent):
    nid = 'lost_on_take_hit'
    desc = "This skill is lost when receiving an attack (it must hit)"
    tag = SkillTags.CUSTOM

    author = 'Lord_Tweed'

    def after_take_strike(self, actions, playback, unit, item, target, mode, attack_info, strike):
        if target and skill_system.check_enemy(unit, target) and strike == Strike.HIT:
            action.do(action.RemoveSkill(unit, self.skill))

class LoseSkillAfterTakeDamage(SkillComponent):
    nid = 'lose_skill_after_take_damage'
    desc = "Lose a skill immediately after an enemy damages you"
    tag = SkillTags.CUSTOM

    def after_take_strike(self, actions, playback, unit, item, target, mode, attack_info, strike):
        for act in actions:
            if isinstance(act, action.ChangeHP) and act.num < 0 and act.unit == unit:
                action.do(action.RemoveSkill(unit, self.skill))
            
class Disvantage(SkillComponent):
    nid = 'disvantage'
    desc = "Unit will attack second even while initiating"
    tag = SkillTags.CUSTOM

    def disvantage(self, unit):
        return True
        
class Damage_Taken(SkillComponent):
    nid = 'Damage_Taken'
    desc = ""
    tag = SkillTags.CUSTOM
    
    def after_take_strike(self, actions, playback, unit, item, target, mode, attack_info, strike):
        stat_changes = {}
        for act in reversed(actions):
            if isinstance(act, action.ChangeHP) and act.num < 0 and act.unit == unit: 
                stat_changes['STR'] = max(0, 5 + act.num)
                #change['STR'] = 3
                action.do(action.ApplyStatChanges(unit, stat_changes, True))
                #self.test.append(change)
    
    #def modify_damage(self, unit, item):
    #    return self.test
    
    '''def after_take_strike(self, actions, playback, unit, item, target, mode, attack_info, strike):
        total_damage_dealt = 0
        playbacks = [p for p in playback if p.nid in (
            'damage_hit', 'damage_crit') and p.attacker == target]
        for p in playbacks:
            total_damage_dealt += p.true_damage
            
        damage = utils.clamp(total_damage_dealt, 0, unit.get_hp())
        test = max(0, 5 - damage)
        unit.get_stat('STR') += test
        unit.get_stat('MAG') += test
    '''    
        #def modify_damage(self, unit, item):
        #    return self.test
        
class EventBeforeCombat(SkillComponent):
    nid = 'event_before_combat'
    desc = 'Calls event on combat start'
    tag = SkillTags.CUSTOM

    expose = ComponentType.Event
    value = ''

    def start_combat(self, playback, unit, item, target, mode):
        game.events.trigger_specific_event(self.value, unit, target, unit.position, {'item': item, 'mode': mode})

class FatalDamage(SkillComponent):
    nid = 'fatal_damage'
    desc = "Skill can deal fatal damage. Use on statuses such as Poison, Bleed, Infection, etc."
    tag = SkillTags.CUSTOM

class FatalBlock(SkillComponent):
    nid = 'fatal_block'
    desc = "This skill should prevent death from fatal damage statuses. Use on statuses such as Legend, Hope for Humanity, etc."
    tag = SkillTags.CUSTOM
    
class CannotUsePhysicalItems(SkillComponent):
    nid = 'cannot_use_physical_items'
    desc = "Unit cannot use or equip physical items"
    tag = SkillTags.CUSTOM

    def available(self, unit, item) -> bool:
        return item_funcs.is_magic(unit, item) or not item.target_enemy
        
class CannotUseItemsOnEnemy(SkillComponent):
    nid = 'cannot_use_items_enemy'
    desc = "Unit cannot use or equip any items that target specifically enemies"
    tag = SkillTags.CUSTOM

    def available(self, unit, item) -> bool:
        return not item.target_enemy
        
class CannotUseItemsOnAlly(SkillComponent):
    nid = 'cannot_use_items_ally'
    desc = "Unit cannot use or equip any items that targets allies"
    tag = SkillTags.CUSTOM

    def available(self, unit, item) -> bool:
        return not item.target_ally
        
class EndstepSkillGain(SkillComponent):
    nid = 'endstep_skill_gain'
    desc = "Grants the designated skill at endstep"
    tag = SkillTags.CUSTOM

    expose = ComponentType.Skill

    def on_endstep(self, actions, playback, unit):
        action.do(action.AddSkill(unit, self.value))